from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

import yaml

from zenconfig.base import BaseConfig, Format


@dataclass
class YAMLFormat(Format):
    indent: int = 2
    sort_keys: bool = True

    @classmethod
    def handles(cls, path: Path) -> bool:
        return path.suffix in {".yml", ".yaml"}

    def load(self, path: Path) -> Dict[str, Any]:
        return yaml.safe_load(path.read_text())

    def dump(self, path: Path, config: Dict[str, Any]) -> None:
        path.write_text(
            yaml.safe_dump(config, indent=self.indent, sort_keys=self.sort_keys)
        )


BaseConfig.register_format(YAMLFormat)



from dataclasses import dataclass
from typing import Any, Dict, Type, TypeVar

from pydantic import BaseModel

from zenconfig.base import BaseConfig, Schema

C = TypeVar("C", bound=BaseModel)


@dataclass
class PydanticSchema(Schema[BaseModel]):
    exclude_unset: bool = False
    exclude_defaults: bool = True

    @classmethod
    def handles(cls, config_class: type) -> bool:
        return issubclass(config_class, BaseModel)

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return cls.parse_obj(cfg)

    def to_dict(self, config: C) -> Dict[str, Any]:
        return config.dict(
            exclude_unset=self.exclude_unset,
            exclude_defaults=self.exclude_defaults,
        )


BaseConfig.register_schema(PydanticSchema)



import sys
from abc import ABC
from typing import ClassVar, Dict

from zenconfig.read import ReadOnlyConfig


class Config(ReadOnlyConfig, ABC):
    FILE_MODE: ClassVar[int] = 0o600

    def save(self) -> None:
        """Save the current config to the file."""
        if not self._path().exists():
            self._path().touch(mode=self.FILE_MODE)
        self._format().dump(self._path(), self._schema().to_dict(self))

    def clear(self) -> None:
        """Remove the config file."""
        kwargs: Dict[str, bool] = {}
        if sys.version_info[:2] != (3, 7):
            kwargs["missing_ok"] = True
        self._path().unlink(**kwargs)



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, ClassVar, Dict, Generic, List, Type, TypeVar, Union


class Format(ABC):
    @classmethod
    @abstractmethod
    def handles(cls, path: Path) -> bool:
        """Return whether the format handles the extension."""

    @abstractmethod
    def load(self, path: Path) -> Dict[str, Any]:
        """Load the configuration file into a dict."""

    @abstractmethod
    def dump(self, path: Path, config: Dict[str, Any]) -> None:
        """Dump in the configuration file."""


C = TypeVar("C")


class Schema(ABC, Generic[C]):
    @classmethod
    @abstractmethod
    def handles(cls, config_class: type) -> bool:
        """Return whether the schema handles the config."""

    @abstractmethod
    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        """Load the schema based on a dict configuration."""

    @abstractmethod
    def to_dict(self, config: Any) -> Dict[str, Any]:
        """Dump the config to dict."""


class BaseConfig(ABC):
    ENV_PATH: ClassVar[str] = "CONFIG"
    PATH: ClassVar[Union[str, None]] = None
    _PATH: ClassVar[Union[Path, None]] = None

    FORMATS: ClassVar[List[Type[Format]]] = []
    FORMAT: ClassVar[Union[Format, None]] = None

    SCHEMAS: ClassVar[List[Type[Schema]]] = []
    SCHEMA: ClassVar[Union[Schema, None]] = None

    @classmethod
    def register_format(cls, format_class: Type[Format]) -> None:
        cls.FORMATS.append(format_class)

    @classmethod
    def register_schema(cls, schema_class: Type[Schema]) -> None:
        cls.SCHEMAS.append(schema_class)

    @classmethod
    def _path(cls) -> Path:
        if cls._PATH:
            return cls._PATH
        found_path: Union[str, None] = None
        if cls.ENV_PATH:
            found_path = os.environ.get(cls.ENV_PATH)
        if not found_path:
            found_path = cls.PATH
        if not found_path:
            raise ValueError(
                f"could not find the config path for config {cls.__qualname__}, tried env variable {cls.ENV_PATH}"
            )
        cls._PATH = Path(found_path).expanduser().absolute()
        return cls._PATH

    @classmethod
    def _format(cls) -> Format:
        if cls.FORMAT:
            return cls.FORMAT
        ext = cls._path().suffix
        path = cls._path()
        for format_class in cls.FORMATS:
            if not format_class.handles(path):
                continue
            cls.FORMAT = format_class()
            return cls.FORMAT
        raise ValueError(
            f"unsupported config file extension {ext} for config {cls.__qualname__}, maybe you are missing an extra"
        )

    @classmethod
    def _schema(cls) -> Schema:
        if cls.SCHEMA:
            return cls.SCHEMA
        for schema_class in cls.SCHEMAS:
            if not schema_class.handles(cls):
                continue
            cls.SCHEMA = schema_class()
            return cls.SCHEMA
        raise ValueError(
            f"could not infer config schema for config {cls.__qualname__}, maybe you are missing an extra"
        )



from dataclasses import asdict, fields, is_dataclass
from typing import Any, Dict, Type, TypeVar

from zenconfig.base import BaseConfig, Schema

C = TypeVar("C")


class DataclassSchema(Schema[C]):
    @classmethod
    def handles(cls, config_class: type) -> bool:
        return is_dataclass(config_class)

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return _load_nested(cls, cfg)

    def to_dict(self, config: C) -> Dict[str, Any]:
        cfg: Dict[str, Any] = {}
        for field in fields(config):
            value = getattr(config, field.name)
            if is_dataclass(field.type):
                value = asdict(value)
            cfg[field.name] = value
        return cfg


BaseConfig.register_schema(DataclassSchema)


def _load_nested(cls: Type[C], cfg: Dict[str, Any]) -> C:
    """Load nested dataclasses."""
    kwargs: Dict[str, Any] = {}
    for field in fields(cls):
        if field.name not in cfg:
            continue
        value = cfg[field.name]
        if is_dataclass(field.type):
            value = _load_nested(field.type, value)
        kwargs[field.name] = value
    return cls(**kwargs)



from abc import ABC
from typing import Type, TypeVar

from zenconfig.base import BaseConfig

C = TypeVar("C", bound="ReadOnlyConfig")


class ReadOnlyConfig(BaseConfig, ABC):
    @classmethod
    def load(cls: Type[C]) -> C:
        return cls._schema().from_dict(cls, cls._format().load(cls._path()))



from typing import Any, Dict, Type, TypeVar

from zenconfig.base import BaseConfig, Schema

C = TypeVar("C", bound=dict)


class DictSchema(Schema[C]):
    @classmethod
    def handles(cls, config_class: type) -> bool:
        return issubclass(config_class, dict)

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return cls(cfg)

    def to_dict(self, config: C) -> Dict[str, Any]:
        return dict(config)


BaseConfig.register_schema(DictSchema)
